@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2022 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -24,6 +24,7 @@
 #define MUPDF_FITZ_DOCUMENT_H
 
 #include "mupdf/fitz/system.h"
+#include "mupdf/fitz/types.h"
 #include "mupdf/fitz/context.h"
 #include "mupdf/fitz/geometry.h"
 #include "mupdf/fitz/device.h"
@@ -32,24 +33,10 @@
 #include "mupdf/fitz/outline.h"
 #include "mupdf/fitz/separation.h"
 
-typedef struct fz_document fz_document;
 typedef struct fz_document_handler fz_document_handler;
 typedef struct fz_page fz_page;
 typedef intptr_t fz_bookmark;
 
-/**
-	Locations within the document are referred to in terms of
-	chapter and page, rather than just a page number. For some
-	documents (such as epub documents with large numbers of pages
-	broken into many chapters) this can make navigation much faster
-	as only the required chapter needs to be decoded at a time.
-*/
-typedef struct
-{
-	int chapter;
-	int page;
-} fz_location;
-
 /**
 	Simple constructor for fz_locations.
 */
@@ -148,6 +135,12 @@ typedef int (fz_document_has_permission_fn)(fz_context *ctx, fz_document *doc, f
 */
 typedef fz_outline *(fz_document_load_outline_fn)(fz_context *ctx, fz_document *doc);
 
+/**
+	Type for a function to be called to obtain an outline iterator
+	for a document. See fz_document_outline_iterator for more information.
+*/
+typedef fz_outline_iterator *(fz_document_outline_iterator_fn)(fz_context *ctx, fz_document *doc);
+
 /**
 	Type for a function to be called to lay
 	out a document. See fz_layout_document for more information.
@@ -157,9 +150,16 @@ typedef void (fz_document_layout_fn)(fz_context *ctx, fz_document *doc, float w,
 /**
 	Type for a function to be called to
 	resolve an internal link to a location (chapter/page number
-	tuple). See fz_resolve_link for more information.
+	tuple). See fz_resolve_link_dest for more information.
 */
-typedef fz_location (fz_document_resolve_link_fn)(fz_context *ctx, fz_document *doc, const char *uri, float *xp, float *yp);
+typedef fz_link_dest (fz_document_resolve_link_dest_fn)(fz_context *ctx, fz_document *doc, const char *uri);
+
+/**
+	Type for a function to be called to
+	create an internal link to a destination (chapter/page/x/y/w/h/zoom/type
+	tuple). See fz_resolve_link_dest for more information.
+*/
+typedef char * (fz_document_format_link_uri_fn)(fz_context *ctx, fz_document *doc, fz_link_dest dest);
 
 /**
 	Type for a function to be called to
@@ -183,11 +183,18 @@ typedef fz_page *(fz_document_load_page_fn)(fz_context *ctx, fz_document *doc, i
 
 /**
 	Type for a function to query
-	a documents metadata. See fz_lookup_metadata for more
+	a document's metadata. See fz_lookup_metadata for more
 	information.
 */
 typedef int (fz_document_lookup_metadata_fn)(fz_context *ctx, fz_document *doc, const char *key, char *buf, int size);
 
+/**
+	Type for a function to set
+	a document's metadata. See fz_set_metadata for more
+	information.
+*/
+typedef int (fz_document_set_metadata_fn)(fz_context *ctx, fz_document *doc, const char *key, const char *value);
+
 /**
 	Return output intent color space if it exists
 */
@@ -278,6 +285,11 @@ typedef int (fz_page_uses_overprint_fn)(fz_context *ctx, fz_page *page);
 */
 typedef fz_link *(fz_page_create_link_fn)(fz_context *ctx, fz_page *page, fz_rect bbox, const char *uri);
 
+/**
+	Type for a function to delete a link on a page.
+*/
+typedef void (fz_page_delete_link_fn)(fz_context *ctx, fz_page *page, fz_link *link);
+
 /**
 	Function type to open a document from a file.
 
@@ -483,6 +495,13 @@ int fz_authenticate_password(fz_context *ctx, fz_document *doc, const char *pass
 */
 fz_outline *fz_load_outline(fz_context *ctx, fz_document *doc);
 
+/**
+	Get an iterator for the document outline.
+
+	Should be freed by fz_drop_outline_iterator.
+*/
+fz_outline_iterator *fz_new_outline_iterator(fz_context *ctx, fz_document *doc);
+
 /**
 	Is the document reflowable.
 
@@ -517,6 +536,21 @@ fz_location fz_lookup_bookmark(fz_context *ctx, fz_document *doc, fz_bookmark ma
 */
 int fz_count_pages(fz_context *ctx, fz_document *doc);
 
+/**
+	Resolve an internal link to a page number, location, and possible viewing parameters.
+
+	Returns location (-1,-1) if the URI cannot be resolved.
+*/
+fz_link_dest fz_resolve_link_dest(fz_context *ctx, fz_document *doc, const char *uri);
+
+/**
+	Format an internal link to a page number, location, and possible viewing parameters,
+	suitable for use with fz_create_link.
+
+	Returns a newly allocated string that the caller must free.
+*/
+char *fz_format_link_uri(fz_context *ctx, fz_document *doc, fz_link_dest dest);
+
 /**
 	Resolve an internal link to a page number.
 
@@ -755,10 +789,17 @@ int fz_lookup_metadata(fz_context *ctx, fz_document *doc, const char *key, char
 #define FZ_META_FORMAT "format"
 #define FZ_META_ENCRYPTION "encryption"
 
-#define FZ_META_INFO_AUTHOR "info:Author"
+#define FZ_META_INFO "info:"
 #define FZ_META_INFO_TITLE "info:Title"
+#define FZ_META_INFO_AUTHOR "info:Author"
+#define FZ_META_INFO_SUBJECT "info:Subject"
+#define FZ_META_INFO_KEYWORDS "info:Keywords"
 #define FZ_META_INFO_CREATOR "info:Creator"
 #define FZ_META_INFO_PRODUCER "info:Producer"
+#define FZ_META_INFO_CREATIONDATE "info:CreationDate"
+#define FZ_META_INFO_MODIFICATIONDATE "info:ModDate"
+
+void fz_set_metadata(fz_context *ctx, fz_document *doc, const char *key, const char *value);
 
 /**
 	Find the output intent colorspace if the document has defined
@@ -789,6 +830,11 @@ int fz_page_uses_overprint(fz_context *ctx, fz_page *page);
 */
 fz_link *fz_create_link(fz_context *ctx, fz_page *page, fz_rect bbox, const char *uri);
 
+/**
+	Delete an existing link on a page.
+*/
+void fz_delete_link(fz_context *ctx, fz_page *page, fz_link *link);
+
 /**
 	Iterates over all opened pages of the document, calling the
 	provided callback for each page for processing. If the callback
@@ -828,6 +874,7 @@ struct fz_page
 	fz_page_separations_fn *separations;
 	fz_page_uses_overprint_fn *overprint;
 	fz_page_create_link_fn *create_link;
+	fz_page_delete_link_fn *delete_link;
 	fz_page **prev, *next; /* linked list of currently open pages */
 };
 
@@ -845,14 +892,17 @@ struct fz_document
 	fz_document_authenticate_password_fn *authenticate_password;
 	fz_document_has_permission_fn *has_permission;
 	fz_document_load_outline_fn *load_outline;
+	fz_document_outline_iterator_fn *outline_iterator;
 	fz_document_layout_fn *layout;
 	fz_document_make_bookmark_fn *make_bookmark;
 	fz_document_lookup_bookmark_fn *lookup_bookmark;
-	fz_document_resolve_link_fn *resolve_link;
+	fz_document_resolve_link_dest_fn *resolve_link_dest;
+	fz_document_format_link_uri_fn *format_link_uri;
 	fz_document_count_chapters_fn *count_chapters;
 	fz_document_count_pages_fn *count_pages;
 	fz_document_load_page_fn *load_page;
 	fz_document_lookup_metadata_fn *lookup_metadata;
+	fz_document_set_metadata_fn *set_metadata;
 	fz_document_output_intent_fn *get_output_intent;
 	fz_document_output_accelerator_fn *output_accelerator;
 	int did_layout;
